#include <Windows.h>
#include <algorithm>
#include <memory>


#include <iostream>
#include <fstream>
#include <sstream>
#include <string>
#include <atlstr.h>
#include <msclr/marshal_cppstd.h>

#include <iterator>

#define __STDC_WANT_LIB_EXT1__ 1
#include <string.h>
#include <stdio.h>

#include <iostream>
#include <bitset>
using namespace std;



// Save the bitmap to a bmp file  
void SaveBitmapToFile(BYTE* pBitmapBits,
    LONG lWidth,
    LONG lHeight,
    WORD wBitsPerPixel,
    const unsigned long& padding_size,
    LPCTSTR lpszFileName)
{
    // Some basic bitmap parameters  
    unsigned long headers_size = sizeof(BITMAPFILEHEADER) +
        sizeof(BITMAPINFOHEADER);

    unsigned long pixel_data_size = lHeight * ((lWidth * (wBitsPerPixel / 8)) + padding_size);

    BITMAPINFOHEADER bmpInfoHeader = { 0 };

    // Set the size  
    bmpInfoHeader.biSize = sizeof(BITMAPINFOHEADER);

    // Bit count  
    bmpInfoHeader.biBitCount = wBitsPerPixel;

    // Use all colors  
    bmpInfoHeader.biClrImportant = 0;

    // Use as many colors according to bits per pixel  
    bmpInfoHeader.biClrUsed = 0;

    // Store as un Compressed  
    bmpInfoHeader.biCompression = BI_RGB;

    // Set the height in pixels  
    bmpInfoHeader.biHeight = lHeight;

    // Width of the Image in pixels  
    bmpInfoHeader.biWidth = lWidth;

    // Default number of planes  
    bmpInfoHeader.biPlanes = 1;

    // Calculate the image size in bytes  
    bmpInfoHeader.biSizeImage = pixel_data_size;

    BITMAPFILEHEADER bfh = { 0 };

    // This value should be values of BM letters i.e 0x4D42  
    // 0x4D = M 042 = B storing in reverse order to match with endian  
    bfh.bfType = 0x4D42;
    //bfh.bfType = 'B'+('M' << 8); 

    // <<8 used to shift M to end  */  

    // Offset to the RGBQUAD  
    bfh.bfOffBits = headers_size;

    // Total size of image including size of headers  
    bfh.bfSize = headers_size + pixel_data_size;

    // Create the file in disk to write  
    HANDLE hFile = CreateFile(lpszFileName,
        GENERIC_WRITE,
        0,
        NULL,
        CREATE_ALWAYS,
        FILE_ATTRIBUTE_NORMAL,
        NULL);

    // Return if error opening file  
    if (!hFile) return;

    DWORD dwWritten = 0;

    // Write the File header  
    WriteFile(hFile,
        &bfh,
        sizeof(bfh),
        &dwWritten,
        NULL);

    // Write the bitmap info header  
    WriteFile(hFile,
        &bmpInfoHeader,
        sizeof(bmpInfoHeader),
        &dwWritten,
        NULL);

    // Write the RGB Data  
    WriteFile(hFile,
        pBitmapBits,
        bmpInfoHeader.biSizeImage,
        &dwWritten,
        NULL);

    // Close the file handle  
    CloseHandle(hFile);
}

BYTE* LoadBMP(int* width, int* height, unsigned long* size, LPCTSTR bmpfile)
{
    BITMAPFILEHEADER bmpheader;
    BITMAPINFOHEADER bmpinfo;
    // value to be used in ReadFile funcs
    DWORD bytesread;
    // open file to read from
    HANDLE file = CreateFile(bmpfile,
        GENERIC_READ,
        FILE_SHARE_READ,
        NULL,
        OPEN_EXISTING,
        FILE_FLAG_SEQUENTIAL_SCAN,
        NULL);
    if (NULL == file)
        return NULL;

    if (ReadFile(file, &bmpheader, sizeof(BITMAPFILEHEADER), &bytesread, NULL) == false)
    {
        CloseHandle(file);
        return NULL;
    }

    // Read bitmap info
    if (ReadFile(file, &bmpinfo, sizeof(BITMAPINFOHEADER), &bytesread, NULL) == false)
    {
        CloseHandle(file);
        return NULL;
    }

    // check if file is actually a bmp
    if (bmpheader.bfType != 'MB')
    {
        CloseHandle(file);
        return NULL;
    }

    // get image measurements
    *width = bmpinfo.biWidth;
    *height = abs(bmpinfo.biHeight);

    // Check if bmp iuncompressed
    if (bmpinfo.biCompression != BI_RGB)
    {
        CloseHandle(file);
        return NULL;
    }

    // Check if we have 24 bit bmp
    if (bmpinfo.biBitCount != 24)
    {
        CloseHandle(file);
        return NULL;
    }

    // create buffer to hold the data
    *size = bmpheader.bfSize - bmpheader.bfOffBits;
    BYTE* Buffer = new BYTE[*size];
    // move file pointer to start of bitmap data
    SetFilePointer(file, bmpheader.bfOffBits, NULL, FILE_BEGIN);
    // read bmp data
    if (ReadFile(file, Buffer, *size, &bytesread, NULL) == false)
    {
        delete[] Buffer;
        CloseHandle(file);
        return NULL;
    }

    // everything successful here: close file and return buffer

    CloseHandle(file);

    return Buffer;
}

std::unique_ptr<BYTE[]> CreateNewBuffer(unsigned long& padding,
    BYTE* pmatrix,
    const int& width,
    const int& height)
{
    padding = (4 - ((width * 3) % 4)) % 4;
    int scanlinebytes = width * 3;
    int total_scanlinebytes = scanlinebytes + padding;
    long newsize = height * total_scanlinebytes;
    std::unique_ptr<BYTE[]> newbuf(new BYTE[newsize]);

    // Fill new array with original buffer, pad remaining with zeros
    std::fill(&newbuf[0], &newbuf[newsize], 0);
    long bufpos = 0;
    long newpos = 0;
    for (int y = 0; y < height; y++)
    {
        for (int x = 0; x < 3 * width; x += 3)
        {
            // Determine positions in original and padded buffers
            bufpos = y * 3 * width + (3 * width - x);
            newpos = (height - y - 1) * total_scanlinebytes + x;

            // Swap R&B, G remains, swap B&R
            newbuf[newpos] = pmatrix[bufpos + 2];
            newbuf[newpos + 1] = pmatrix[bufpos + 1];
            newbuf[newpos + 2] = pmatrix[bufpos];
        }
    }

   

    return newbuf;


}
  
static void convert_row(byte* buf, int* pos, int** bitplane, int p) {

    static const unsigned char colours[16][3] = { // An array of colour values
        //BBB  GGG  RRR
        { 255, 255, 255 }, /* White    */
        {   0,  48,  64 }, /* Index  1 */
        {   0,  72, 112 }, /* Index  2 */
        {  56, 104, 168 }, /* Index  3 */
        {   0, 152, 216 }, /* Index  4 */
        {   8, 216, 248 }, /* Index  5 */
        { 255, 255, 255 }, /* Index  6 */
        { 208, 208, 208 }, /* Index  7 */
        { 160, 160, 160 }, /* Index  8 */
        { 112, 120, 120 }, /* Index  9 */
        {   0,   0,  72 }, /* Index 10 */
        {   8,   8, 160 }, /* Index 11 */
        {  16,  16, 248 }, /* Index 12 */
        {   8, 104, 248 }, /* Index 13 */
        {   0, 248,   0 }, /* Index 14 */
        {   0,   0,   0 }  /* Index 15 */
    };

    for (int N = 7; N >= 0; N--) {
        std::bitset<4> bs;
        bs.set(0, (bitplane[0][p] >> N) & 1);
        bs.set(1, (bitplane[1][p] >> N) & 1);
        bs.set(2, (bitplane[2][p] >> N) & 1);
        bs.set(3, (bitplane[3][p] >> N) & 1);
        unsigned long Index = bs.to_ulong();
        int c = *pos;
        buf[c++] = colours[Index][0];   // Blue
        buf[c++] = colours[Index][1];   // Green
        buf[c++] = colours[Index][2];   // Red
        *pos = c;
    }
}


int main() {

    CString File_Path;
    char sztmp[1024];

    cin.clear();
    cout << "Enter Path to extract sprite graphics to: " << endl;
    cin.getline(sztmp, sizeof(sztmp), '\n');
    File_Path = sztmp;



    char oData[0x2000]; // The Array
    ifstream inFileB;
    inFileB.open("C:\\Users\\Chris\\Desktop\\DK\\Emulators ad Debuggers\\zsnesw150\\dkc1.sfc");
    inFileB.seekg(0x347519 + 10);
    inFileB.read(oData, 0x2000);
    inFileB.close();

    int E = 0;

    // there are only 26 letters, so we have 26 iterations
    for (int X = 0; X <= 25; X++) {

        // These allocations are kept the same
        // but there are a few ways to make them a bit simpler.

        // Letter K -- Top Left Tile
        int* Bitplane0_Top_Left_ROW = new int[8];
        int* Bitplane1_Top_Left_ROW = new int[8];
        int* Bitplane2_Top_Left_ROW = new int[8];
        int* Bitplane3_Top_Left_ROW = new int[8];

        // Letter K -- Top Right Tile
        int* Bitplane0_Top_Right_ROW = new int[8];
        int* Bitplane1_Top_Right_ROW = new int[8];
        int* Bitplane2_Top_Right_ROW = new int[8];
        int* Bitplane3_Top_Right_ROW = new int[8];

        // Letter K -- Bottom Left Tile
        int* Bitplane0_Bottom_Left_ROW = new int[8];
        int* Bitplane1_Bottom_Left_ROW = new int[8];
        int* Bitplane2_Bottom_Left_ROW = new int[8];
        int* Bitplane3_Bottom_Left_ROW = new int[8];

        // Letter K -- Bottom Right Tile
        int* Bitplane0_Bottom_Right_ROW = new int[8];
        int* Bitplane1_Bottom_Right_ROW = new int[8];
        int* Bitplane2_Bottom_Right_ROW = new int[8];
        int* Bitplane3_Bottom_Right_ROW = new int[8];

        // the assignments can be reduced to a pattern,
        // and iterated with a loop.
        // Can they be reduced further?
        for (int i = 0; i < 8; i++) {
            int offset = E + 2 * (7 - i);
            Bitplane0_Top_Left_ROW[i] = oData[offset];
            Bitplane1_Top_Left_ROW[i] = oData[offset + 1];
            Bitplane2_Top_Left_ROW[i] = oData[offset + 16];
            Bitplane3_Top_Left_ROW[i] = oData[offset + 17];
            Bitplane0_Top_Right_ROW[i] = oData[offset + 32];
            Bitplane1_Top_Right_ROW[i] = oData[offset + 33];
            Bitplane2_Top_Right_ROW[i] = oData[offset + 48];
            Bitplane3_Top_Right_ROW[i] = oData[offset + 49];
            Bitplane0_Bottom_Left_ROW[i] = oData[offset + 64];
            Bitplane1_Bottom_Left_ROW[i] = oData[offset + 65];
            Bitplane2_Bottom_Left_ROW[i] = oData[offset + 80];
            Bitplane3_Bottom_Left_ROW[i] = oData[offset + 81];
            Bitplane0_Bottom_Right_ROW[i] = oData[offset + 96];
            Bitplane1_Bottom_Right_ROW[i] = oData[offset + 97];
            Bitplane2_Bottom_Right_ROW[i] = oData[offset + 112];
            Bitplane3_Bottom_Right_ROW[i] = oData[offset + 113];
        }

        // E is set here for the next iteration
        E = E + 128 + 10;

        // 16 wide, 16 tall, 3 bytes per pixel = 16*16*3 = 768
        BYTE* buf = new BYTE[3264];
        int c = 0;

        // If we group the bitmap arrays together,
        // it makes it a bit easier to pass them to a function
        int* bitplane[][4] = {
            { Bitplane0_Top_Left_ROW,     Bitplane1_Top_Left_ROW,     Bitplane2_Top_Left_ROW,     Bitplane3_Top_Left_ROW     },
            { Bitplane0_Top_Right_ROW,    Bitplane1_Top_Right_ROW,    Bitplane2_Top_Right_ROW,    Bitplane3_Top_Right_ROW    },
            { Bitplane0_Bottom_Left_ROW,  Bitplane1_Bottom_Left_ROW,  Bitplane2_Bottom_Left_ROW,  Bitplane3_Bottom_Left_ROW  },
            { Bitplane0_Bottom_Right_ROW, Bitplane1_Bottom_Right_ROW, Bitplane2_Bottom_Right_ROW, Bitplane3_Bottom_Right_ROW }
        };

        // convert each tile one row at a time
        for (int p = 0; p < 8; p++) {
            convert_row(buf, &c, bitplane[2], p); // Bottom Left
            convert_row(buf, &c, bitplane[3], p); // Bottom Right
        }

        for (int p = 0; p < 8; p++) {
            convert_row(buf, &c, bitplane[0], p); // Top Left
            convert_row(buf, &c, bitplane[1], p); // Top Right
        }



        // Generate a filename and create the bitmap
       // The destination path
        SaveBitmapToFile((BYTE*)buf, 16, 16, 24, 0, File_Path + X.ToString() + "_test.bmp");

        
        // Delete allocations
        delete[] buf;
        delete[] Bitplane0_Top_Left_ROW;
        delete[] Bitplane1_Top_Left_ROW;
        delete[] Bitplane2_Top_Left_ROW;
        delete[] Bitplane3_Top_Left_ROW;
        delete[] Bitplane0_Top_Right_ROW;
        delete[] Bitplane1_Top_Right_ROW;
        delete[] Bitplane2_Top_Right_ROW;
        delete[] Bitplane3_Top_Right_ROW;
        delete[] Bitplane0_Bottom_Left_ROW;
        delete[] Bitplane1_Bottom_Left_ROW;
        delete[] Bitplane2_Bottom_Left_ROW;
        delete[] Bitplane3_Bottom_Left_ROW;
        delete[] Bitplane0_Bottom_Right_ROW;
        delete[] Bitplane1_Bottom_Right_ROW;
        delete[] Bitplane2_Bottom_Right_ROW;
        delete[] Bitplane3_Bottom_Right_ROW;

    }

    
    return 0;

}






